// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"
#include "objective.h"

#include "dco.hpp"

template<typename T>
T fit(T p, VT<T> xobs, VT<T> yobs);

/// error analysis by symbolic differentiation
template<typename T>
VT<T> dp_dobs_sym(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // result of differentiation of first-order optimality condition
  return -dde_dobs_dp(p,xobs,yobs)/dde_dp_dp(p,xobs,yobs);
}

/// error analysis by central finite differences
template<typename T>
VT<T> dp_dobs_cfd(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // return value
  VT<T> dpdobs(2*nobs);
  int i=0, j=0;
  while (i<nobs) {
    T xobs_i_save=xobs(i), yobs_i_save=yobs(i), pp, pm; 
    // perturbing approx. midpoint of mantissa
    T delta=eps;
    if (xobs(i)!=0) {
      delta=delta*fabs(xobs(i));
    }
    // central finite difference quotient
    xobs(i)=xobs(i)+delta;
    pp=fit(p,xobs,yobs);
    xobs(i)=xobs_i_save-delta;
    pm=fit(p,xobs,yobs);
    xobs(i)=xobs_i_save;
    dpdobs(j)=(pp-pm)/(delta+delta); j=j+1;
    // perturbing approx. midpoint of mantissa
    delta=eps;
    if (yobs(i)!=0) {
      delta=delta*fabs(yobs(i));
    }
    // central finite difference quotient
    yobs(i)=yobs(i)+delta;
    pp=fit(p,xobs,yobs);
    yobs(i)=yobs_i_save-delta;
    pm=fit(p,xobs,yobs);
    yobs(i)=yobs_i_save;
    dpdobs(j)=(pp-pm)/(delta+delta); j=j+1;
    i=i+1;
  }
  return dpdobs;
}

/// error analysis by tangent AD
template<typename T>
VT<T> dp_dobs_tad(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // tangent type
  using TT=dco::gt1s<T>::type;
  // activation
  TT p_t;
  VT<TT> xobs_t(nobs), yobs_t(nobs);
  int i=0;
  while (i<nobs) {
    xobs_t(i)=xobs(i); 
    yobs_t(i)=yobs(i);
    i=i+1;
  }
  // return value
  VT<T> dpdobs(2*nobs);
  int j=0;
  i=0;
  while (i<nobs) {
    // tangent in standard basis direction of current point of observation
    p_t=p; // reset model parameter to its initial value
    // seed
    dco::derivative(xobs_t(i))=1;
    // overloaded parameter fitting
    p_t=fit(p_t,xobs_t,yobs_t); // all arguments of same type
    // harvest
    dpdobs(j)=dco::derivative(p_t); j=j+1;
    // enable correct next seed
    dco::derivative(xobs_t(i))=0; 
    // tangent in standard basis direction of current observed value
    p_t=p; // reset model parameter to its initial value
    // seed
    dco::derivative(yobs_t(i))=1;
    // overloaded parameter fitting
    p_t=fit(p_t,xobs_t,yobs_t);
    // harvest
    dpdobs(j)=dco::derivative(p_t); j=j+1;
    // enable correct next seed
    dco::derivative(yobs_t(i))=0;
    i=i+1;
  }
  return dpdobs;
}

/// error analysis by adjoint AD
template<typename T>
VT<T> dp_dobs_aad(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // adjoint type
  using AT=dco::ga1s<T>::type;
  // tape
  dco::smart_tape_ptr_t<dco::mode<AT>> tape;
  // activation and registration
  AT p_a=p;
  VT<AT> xobs_a(nobs), yobs_a(nobs);
  int i=0;
  while (i<nobs) {
    xobs_a(i)=xobs(i); 
    tape->register_variable(xobs_a(i));
    yobs_a(i)=yobs(i);
    tape->register_variable(yobs_a(i));
    i=i+1;
  }
  // overloaded parameter fitting
  p_a=fit(p_a,xobs_a,yobs_a);
  // seed
  dco::derivative(p_a)=1;
  // interpretation of tape
  tape->interpret_adjoint();
  // harvest
  VT<T> dpdobs(2*nobs);
  int j=0;
  i=0;
  while (i<nobs) {
    dpdobs(j)=dco::derivative(xobs_a(i)); j=j+1; 
    dpdobs(j)=dco::derivative(yobs_a(i)); j=j+1; 
    i=i+1;
  }
  return dpdobs;
}
