// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"

// user code: right-hand side of explicit ordinary differential equation
template<typename T> 
T g(T p, T x, T y); 

/// residual of implicit Euler equation
template<typename T> 
T r(T p, T x, T y, T y_prev, T delta_x) { 
  // conditions
  assert(x>=0);
  assert(delta_x>0);
  // residual
  return y-y_prev-delta_x*g(p,x,y);
}

/// implicit Euler system solved by bisection
template<typename T>
T bisection(T p, T x, T y, T delta_x) {
  // conditions
  assert(x>=0);
  assert(delta_x>0);
  // search interval (smarter search required in general)
  T y_prev=y, y_up=y+2; 
  y=y-2;
  // iterative bisection to decrease width of search interval
  while (fabs(y-y_up)>eps) {
    // midpoint of search interval
    T y_mid=(y_up+y)/2;
    // residual at lower bound
    T r_=r(p,x,y,y_prev,delta_x), 
    // residual at midpoint
    r_mid=r(p,x,y_mid,y_prev,delta_x);
    // collapse search interval if root found
    if (fabs(r_mid)<eps) {
      y=y_up=y_mid;
    }
    // continue bisection if root not found
    else if ((r_<=0&&r_mid<=0)||(r_>=0&&r_mid>=0)) {
      // search in right half of sign switches there
      y=y_mid;
    } 
    else {
      // search in left half otherwise
      y_up=y_mid;
    } 
  }
  return y;
}

/// implicit Euler simulation of differential model 
template<typename T>
T f(T p, T x) {
  // condition
  assert(x>=0);
  // integration step size
  T delta_x=x/m;
  // initial position fixed to zero
  x=0;
  // initial state fixed to zero
  T y=0;
  int i=0;
  while (i<m) {
    // implicit Euler step
    y=bisection(p,x,y,delta_x); // root of implicit Euler equation
    x=x+delta_x; 
    i=i+1;
  } 
  return y;
}
