// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"

// user code: right-hand side of explicit ordinary differential equation
template<typename T> 
T g(T p, T x, T y);

// user code: derivative of right-hand side of explicit ODE wrt. state
template<typename T> 
T dg_dy(T p, T x, T y);

/// residual of implicit Euler equation
template<typename T> 
T r(T p, T x, T y, T y_prev, T delta_x) { 
  // conditions
  assert(x>=0);
  assert(delta_x>0);
  // residual
  return y-y_prev-delta_x*g(p,x,y);
}

/// derivative of residual implicit Euler system
template<typename T> 
T dr_dy(T p, T x, T y, T delta_x) { 
  // conditions
  assert(x>=0);
  assert(delta_x>0);
  // derivative of residual 
  return 1-delta_x*dg_dy(p,x,y);
}

/// implicit Euler system solved by Newton method
template<typename T>
T newton(T p, T x, T y, T delta_x) {
  // conditions
  assert(x>=0);
  assert(delta_x>0);
  // state computed by previous implicit Euler step
  T y_prev=y;
  do {
    // Newton step
    y=y-r(p,x,y,y_prev,delta_x)/dr_dy(p,x,y,delta_x);
  } while (fabs(r(p,x,y,y_prev,delta_x))>eps);
  return y;
}

/// implicit Euler simulation of differential model 
template<typename T>
T f(T p, T x) {
  // condition
  assert(x>=0);
  // integration step size 
  T delta_x=x/m;
  // initial position fixed to zero
  x=0;
  // initial state fixed to zero
  T y=0;
  int i=0;
  while (i<m) {
    // implicit Euler step
    y=newton(p,x,y,delta_x); // root of implicit Euler equation
    x=x+delta_x; 
    i=i+1;
  } 
  return y;
}
