// naumann@stce.rwth-aachen.de

#pragma once

#include "config.h"

#include "dco.hpp"

// model
template<typename T>
T f(T p, T x);

/// objective is least squares error of model predictions relative to observations
template<typename T>
T e(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  T e=0;
  int i=0;
  while (i<nobs) {
    e=e+pow(f(p,xobs(i))-yobs(i),2);
    i=i+1;
  } 
  return e;
}

/// first derivative of objective wrt. model parameter by tangent AD
template<typename T>
T de_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // tangent type
  using TT=dco::gt1s<T>::type;
  // activation
  TT p_t=p, e_t;
  // seeding derivative of model parameter
  dco::derivative(p_t)=1;
  // overloaded computation of objective (explicit template instantiation required due to passive xobs and yobs)
  e_t=e<TT>(p_t,xobs,yobs);
  // harvesting derivative of objective wrt. model parameter
  return dco::derivative(e_t);
}

/// derivative of derivative of objective wrt. model parameter wrt. model parameter by finite differences
template<typename T>
T dde_dp_dp_cfd(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // perturbing approx. midpoint of mantissa
  T delta_p=eps;
  if (p!=0) {
    delta_p=delta_p*fabs(p);
  }
  // central finite difference of first derivative
  return (de_dp(p+delta_p,xobs,yobs)-de_dp(p-delta_p,xobs,yobs))/(delta_p+delta_p);
}

/// derivative of derivative of objective wrt. model parameter wrt. model parameter by tangent AD
template<typename T>
T dde_dp_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // tangent type
  using TT=dco::gt1s<T>::type;
  // activation
  TT p_t=p, dedp_t;
  // seeding derivative of model parameter
  dco::derivative(p_t)=1;
  // overloaded computation of derivative of objective wrt. model parameter
  dedp_t=de_dp<TT>(p_t,xobs,yobs);
  // harvesting derivative of derivative of objective wrt. model parameter wrt. model parameter
  return dco::derivative(dedp_t);
}

/// derivative of objective wrt. observations by adjoint AD
template<typename T>
VT<T> de_dobs(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // adjoint type
  using AT=dco::ga1s<T>::type;
  // smart pointer to tape
  dco::smart_tape_ptr_t<dco::mode<AT>> tape;
  // activation
  AT e_a;
  VT<AT> xobs_a(nobs), yobs_a(nobs);
  int i=0;
  while (i<nobs) {
    xobs_a(i)=xobs(i); 
    yobs_a(i)=yobs(i);
    i=i+1;
  }
  // registration of observations with tape
  i=0;
  while (i<nobs) {
    tape->register_variable(xobs_a(i));
    tape->register_variable(yobs_a(i));
    i=i+1;
  }
  // overloaded computation of objective records tape
  e_a=e<AT>(p,xobs_a,yobs_a);
  // seeding derivative of objective
  dco::derivative(e_a)=1;
  // interpretation of tape
  tape->interpret_adjoint();
  // harvesting derivative of objective wrt. obserations
  VT<T> de_dobs(2*nobs); 
  int j=0;
  i=0;
  while (i<nobs) {
    de_dobs(j)=dco::derivative(xobs_a(i)); j=j+1;
    de_dobs(j)=dco::derivative(yobs_a(i)); j=j+1; 
    i=i+1;
  }
  return de_dobs;
}

/// derivative of derivative of objective wrt. observations wrt. model parameter by tangent AD
template<typename T>
VT<T> dde_dobs_dp(T p, VT<T> xobs, VT<T> yobs) {
  // conditions
  assert(xobs.size()==nobs);
  assert(xobs.size()==yobs.size());
  // tangent type
  using TT=dco::gt1s<T>::type;
  // activation
  TT p_t=p;
  VT<TT> dedobs_t(nobs);
  // seeding derivative of model parameter
  dco::derivative(p_t)=1;
  // overloaded computation of derivative of objective wrt. observations
  dedobs_t=de_dobs<TT>(p_t,xobs,yobs);
  // harvesting derivative of derivative of objective wrt. observations wrt. model parameter

  VT<T> dde_dobs_dp(2*nobs); 
  int i=0;
  while (i<2*nobs) {
    dde_dobs_dp(i)=dco::derivative(dedobs_t(i));
    i=i+1;
  }
  return dde_dobs_dp;
}
